#    (C) Copyright 2015-2022 Sei Lisa. All rights reserved.
#
#    This file is part of LSL PyOptimizer.
#
#    LSL PyOptimizer is free software: you can redistribute it and/or
#    modify it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    LSL PyOptimizer is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with LSL PyOptimizer. If not, see <http://www.gnu.org/licenses/>.

# Extra functions that have predictable return values for certain arguments.

from .lslcommon import Key, Vector #, Quaternion
from .lslbasefuncs import ELSLCantCompute, fi,ff,fs,fk,v2f,q2f,fl, \
  NULL_KEY, ZERO_VECTOR, ZERO_ROTATION, \
  TOUCH_INVALID_TEXCOORD, cond
#from strutil import unicode
ff, q2f  # keep pyflakes happy as these are not used

GetEnvSettings = frozenset({'agent_limit', 'dynamic_pathfinding', 'estate_id'
    , 'estate_name', 'frame_number', 'region_cpu_ratio', 'region_idle'
    , 'region_product_name', 'region_product_sku', 'region_start_time'
    , 'sim_channel', 'sim_version', 'simulator_hostname'
    , 'region_max_prims'  # <http://wiki.secondlife.com/wiki/Release_Notes/Second_Life_RC_Magnum/16#16.11.02.321369>
    , 'region_object_bonus'  # <http://wiki.secondlife.com/wiki/Release_Notes/Second_Life_RC_Magnum/16#16.12.03.322072>
    , 'whisper_range', 'chat_range', 'shout_range'  #? (jun 2020)
    , 'agent_limit_max', 'agent_reserved', 'agent_underserved'  # <https://releasenotes.secondlife.com/simulator/2022-04-21.571166.html>
    })

# Values valid for llGetVisualDetails parameters
GVD_ValidValues = \
    { 33, "height"
    , 38, "torso_length"
    , 80, "male"
    , 198, "heel_height"
    , 503, "platform_height"
    , 616, "shoe_height"
    , 692, "leg_length"
    , 693, "arm_length"
    , 756, "neck_length"
    , 814, "waist_height"
    , 842, "hip_length"
    , 11001, "hover"
    }

xp_error_messages = {
    -1:u'unknown error id',
    0:u'no error', 1:u'exceeded throttle', 2:u'experiences are disabled',
    3:u'invalid parameters', 4:u'operation not permitted',
    5:u'script not associated with an experience', 6:u'not found',
    7:u'invalid experience', 8:u'experience is disabled',
    9:u'experience is suspended', 10:u'unknown error',
    11:u'experience data quota exceeded',
    12:u'key-value store is disabled',
    13:u'key-value store communication failed', 14:u'key doesn\'t exist',
    15:u'retry update', 16:u'experience content rating too high',
    17:u'not allowed to run in current location',
    18:u'experience permissions request timed out'
}

valid_inventory_kinds = frozenset({0, 1, 3, 5, 6, 7, 10, 13, 20, 21})

def llCloud(v):
    v = v2f(v)
    return 0.0

def llAvatarOnLinkSitTarget(link):
    link = fi(link)
    if link > 255 or link == -2147483648:
        return Key(NULL_KEY)
    raise ELSLCantCompute

# llClearPrimMedia always has side effects (emits errors for every face that
# is not supported)
def llClearLinkMedia(link, face):
    if link > 255 or link == -2147483648:
        return 0
    raise ELSLCantCompute

def llDetectedGrab(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'grab' in evsym):
        raise ELSLCantCompute
    return ZERO_VECTOR

def llDetectedGroup(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'detect' in evsym):
        raise ELSLCantCompute
    return 0

def llDetectedKey(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'detect' in evsym):
        raise ELSLCantCompute
    return Key(NULL_KEY)

def llDetectedLinkNumber(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'detect' in evsym):
        raise ELSLCantCompute
    return 0

def llDetectedName(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'detect' in evsym):
        raise ELSLCantCompute
    return NULL_KEY

def llDetectedOwner(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'detect' in evsym):
        raise ELSLCantCompute
    return Key(NULL_KEY)

def llDetectedPos(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'detect' in evsym):
        raise ELSLCantCompute
    return ZERO_VECTOR

def llDetectedRot(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'detect' in evsym):
        raise ELSLCantCompute
    return ZERO_ROTATION

def llDetectedTouchBinormal(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'touch' in evsym):
        raise ELSLCantCompute
    return ZERO_VECTOR

def llDetectedTouchFace(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'touch' in evsym):
        raise ELSLCantCompute
    return -1 if 'detect' in evsym and 0 <= idx <= 15 else 0

def llDetectedTouchNormal(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'touch' in evsym):
        raise ELSLCantCompute
    return ZERO_VECTOR

def llDetectedTouchPos(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'touch' in evsym):
        raise ELSLCantCompute
    return ZERO_VECTOR

def llDetectedTouchST(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'detect' in evsym):
        # In detection events that are not touch events, it returns
        # TOUCH_INVALID_TEXCOORD if idx < num, else ZERO_VECTOR,
        # but we only know that num >= 1.
        if idx == 0 and evsym is not None and 'touch' not in evsym:
            # index 0 always exists, so we know the result
            return TOUCH_INVALID_TEXCOORD
        raise ELSLCantCompute
    return ZERO_VECTOR

def llDetectedTouchUV(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'detect' in evsym):
        # In detection events that are not touch events, it returns
        # TOUCH_INVALID_TEXCOORD if idx < num, else ZERO_VECTOR,
        # but we only know that num >= 1.
        if idx == 0 and evsym is not None and 'touch' not in evsym:
            # index 0 always exists, so we know the result
            return TOUCH_INVALID_TEXCOORD
        raise ELSLCantCompute
    return ZERO_VECTOR

def llDetectedType(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'detect' in evsym):
        raise ELSLCantCompute
    return 0

def llDetectedVel(idx, evsym=None):
    idx = fi(idx)
    if 0 <= idx <= 15 and (evsym is None or 'detect' in evsym):
        raise ELSLCantCompute
    return ZERO_VECTOR

def llEdgeOfWorld(v1, v2):
    v1 = v2f(v1)
    v2 = v2f(v2)
    if v2[0] == v2[1] == 0:
        return 1
    raise ELSLCantCompute

def llGetAgentInfo(id):
    id = fk(id)
    if not cond(id):
        return 0
    raise ELSLCantCompute

def llGetAgentLanguage(id):
    id = fk(id)
    if not cond(id):
        return u''
    raise ELSLCantCompute

def llGetAgentList(scope, options):
    scope = fi(scope)
    options = fl(options)
    if scope not in (1, 2, 4):
        return [u'INVALID_SCOPE']
    raise ELSLCantCompute

def llGetAgentSize(id):
    id = fk(id)
    if not cond(id):
        return ZERO_VECTOR
    raise ELSLCantCompute

def llGetAlpha(face):
    face = fi(face)
    if face > 8:
        return 1.0
    # Negative face numbers return (float)llGetNumberOfSides(), which isn't
    # computable.
    raise ELSLCantCompute

def llGetAnimation(id):
    id = fk(id)
    if not cond(id):
        return u''
    raise ELSLCantCompute

def llGetAnimationList(id):
    id = fk(id)
    if not cond(id):
        return []
    raise ELSLCantCompute

def llGetAttachedList(id):
    id = fk(id)
    if not cond(id):
        return [u'NOT FOUND']
    raise ELSLCantCompute

def llGetBoundingBox(id):
    id = fk(id)
    if not cond(id):
        return []
    raise ELSLCantCompute

def llGetColor(face):
    face = fi(face)
    if face > 8:
        return Vector((1.,1.,1.))
    # Returns the average colour when negative (can't be computed)
    raise ELSLCantCompute

def llGetDisplayName(id):
    id = fk(id)
    if not cond(id):
        return u''
    raise ELSLCantCompute

def llGetEnv(s):
    s = fs(s)
    if s not in GetEnvSettings:
        return u""
    raise ELSLCantCompute

def llGetExperienceErrorMessage(errno):
    errno = fi(errno)
    if errno < -1 or errno > 18:
        errno = -1
    return xp_error_messages[errno]

def llGetExperienceList(id):
    id = fk(id)
    # This function is not implemented and always returns empty list
    return []

def llGetHTTPHeader(id, s):
    id = fk(id)
    s = fs(s)
    if not cond(id):
        return u''
    raise ELSLCantCompute

def llGetInventoryKey(s):
    s = fs(s)
    if s == u'':
        return Key(NULL_KEY)
    raise ELSLCantCompute

def llGetInventoryName(kind, index):
    kind = fi(kind)
    index = fi(index)
    if kind != -1 and kind not in valid_inventory_kinds or index < 0:
        return u''
    raise ELSLCantCompute

def llGetInventoryNumber(kind):
    kind = fi(kind)
    if kind != -1 and kind not in valid_inventory_kinds:
        return 0
    raise ELSLCantCompute

def llGetInventoryPermMask(item, category):
    item = fs(item)
    category = fi(category)
    if category < 0 or category > 4 or item == u'':
        return 0
    raise ELSLCantCompute

def llGetLinkName(link):
    link = fi(link)
    if link < -2147483646 or link > 256:
        return NULL_KEY
    raise ELSLCantCompute

def llGetObjectLinkKey(id, link):
    # TODO: Investigate behaviour with invalid key, invalid link etc.
    raise ELSLCantCompute

def llGetOwnerKey(id):
    id = fk(id)
    if not cond(id):
        return Key(NULL_KEY)
    raise ELSLCantCompute

def llGetStatus(mask):
    # Leave out STATUS_CAST_SHADOWS from the flags to check.
    if (mask & ~0x200) == 0:
        return 0
    raise ELSLCantCompute

def llGetVisualParams(id, params):
    id = fk(id)
    params = fl(params)
    if not cond(id) or params.nt == 'CONST' and not params.ch:
        return []
# TODO: This needs to check whether the list is a literal, and whether the
# elements are constant.
#    if params.nt == 'LIST':
#    for i in params:
#        if (i.lower() if type(i) == unicode else i) in GVD_ValidValues:
#            raise ELSLCantCompute
#    return [u""] * len(params)
    raise ELSLCantCompute

# TODO: Add more predictable functions.



#    (C) Copyright 2015-2022 Sei Lisa. All rights reserved.
#
#    This file is part of LSL PyOptimizer.
#
#    LSL PyOptimizer is free software: you can redistribute it and/or
#    modify it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    LSL PyOptimizer is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with LSL PyOptimizer. If not, see <http://www.gnu.org/licenses/>.

# Put all LSL functions together in one single module

from .lslbasefuncs import *
from .lsljson import *
from .lslextrafuncs import *
lslbasefuncs_used



#    (C) Copyright 2015-2022 Sei Lisa. All rights reserved.
#
#    This file is part of LSL PyOptimizer.
#
#    LSL PyOptimizer is free software: you can redistribute it and/or
#    modify it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    LSL PyOptimizer is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with LSL PyOptimizer. If not, see <http://www.gnu.org/licenses/>.

# Classes, functions and variables for use of all modules.

import sys
from .strutil import *
strutil_used

_exclusions = frozenset({'nt','t','name','value','ch', 'X','SEF'})

# Node Record type. Used for AST nodes.
class nr(object):
    nt  = None   # node type
    t   = None   # LSL type
    ch  = None   # children
    SEF = False  # Side Effect-Free flag
    def __init__(self, **kwargs):
        for k in kwargs:
            setattr(self, k, kwargs[k])

    def copy(self):
        new = nr()
        for k, v in self.__dict__.items():
            setattr(new, k, v)
        return new

    # Debug output

    def __str__(self, indent = 0):
        spaces = ' ' * (4 * indent)
        s = '\n{sp}{{ nt:{nr.nt}\n{sp}  ,t:{nr.t}'.format(sp=spaces, nr=self)
        if hasattr(self, 'name'):
            s += '\n{sp}  ,name:{nr.name}'.format(sp=spaces, nr=self)
        if hasattr(self, 'value'):
            s += '\n{sp}  ,value:{v}'.format(sp=spaces, v=repr(self.value))
        for k in sorted(self.__dict__):
            if k not in _exclusions:
                v = self.__dict__[k]
                s += '\n{sp}  ,{k}:{v}'.format(sp=spaces, k=k, v=repr(v))
        if self.ch is not None:
            if self.ch:
                s += '\n{sp}  ,ch:['.format(sp=spaces)
                isFirst = True
                for v in self.ch:
                    if not isFirst:
                        s += ','
                    isFirst = False
                    s += v.__str__(indent + 1)
                s += '\n{sp}  ]'.format(sp=spaces)
            else:
                s += '\n{sp}  ,ch:[]'.format(sp=spaces)
        s += '\n{sp}}}'.format(sp=spaces)
        return s if indent > 0 else s[1:]  # remove leading \n at level 0

# These types just wrap the Python types to make type() work on them.
# There are no ops defined on them or anything.

class Key(unicode):
    def __repr__(self):
        return 'Key(' + super(Key, self).__repr__() + ')'

class Vector(tuple):
    def __repr__(self):
        return 'Vector(' + super(Vector, self).__repr__() + ')'

class Quaternion(tuple):
    def __repr__(self):
        return 'Quaternion(' + super(Quaternion, self).__repr__() + ')'

# Recognized: 3763, 6466, 6495
# BUG-3763 affected llXorBase64 (see lslbasefuncs.py). It's been fixed.
# BUG-6466 is about some valid numbers in JSON not being accepted. It's fixed.
# BUG-6495 is about "]" inside a JSON string closing an array. It's NOT FIXED.
Bugs = set([6495])

# LSO is generally not supported, but some calculations can handle it. The
# main intention of this setting is for the code to serve as documentation
# of how LSO's behaviour differs from Mono's in the fine details.
LSO = False

# Set to True by lslcalc's main
IsCalc = False

DataPath = ''

# Language

# These are hardcoded because additions or modifications imply
# important changes to the code anyway.
types = frozenset({'integer','float','string','key','vector',
    'quaternion','rotation','list'})

# Conversion of LSL types to Python types and vice versa.

PythonType2LSL = {int: 'integer', float: 'float',
    unicode: 'string', Key: 'key', Vector: 'vector',
    Quaternion: 'rotation', list: 'list'}

LSLType2Python = {'integer':int, 'float':float,
    'string':unicode, 'key':Key, 'vector':Vector,
    'rotation':Quaternion, 'list':list}

LSLTypeDefaults = {'integer':0, 'float':0.0, 'string':u'', 'key':Key(u''),
    'vector':Vector((0.,0.,0.)), 'rotation':Quaternion((0.,0.,0.,1.)),
    'list':[]}

def warning(txt):
    assert type(txt) == unicode
    sys.stderr.write(u"WARNING: " + txt + u"\n")



from lummao import *


class Script(BaseLSLScript):
    gTestsPassed: int
    gTestsFailed: int
    gInteger: int
    gFloat: float
    gString: str
    gVector: Vector
    gRot: Quaternion
    gList: list
    gCallOrder: list

    def __init__(self):
        super().__init__()
        self.gTestsPassed = 0
        self.gTestsFailed = 0
        self.gInteger = 5
        self.gFloat = 1.5
        self.gString = "foo"
        self.gVector = Vector((1.0, 2.0, 3.0))
        self.gRot = Quaternion((1.0, 2.0, 3.0, 4.0))
        self.gList = [1, 2, 3]
        self.gCallOrder = []

    async def testPassed(self, _description: str, _actual: str, _expected: str) -> None:
        self.gTestsPassed += 1

    async def testFailed(self, _description: str, _actual: str, _expected: str) -> None:
        self.gTestsFailed += 1
        print(radd(")", radd(_expected, radd(" expected ", radd(_actual, radd(" (", radd(_description, "FAILED!: ")))))))
        await self.builtin_funcs.llOwnerSay(typecast(rdiv(0, 0), str))

    async def ensureTrue(self, _description: str, _actual: int) -> None:
        if cond(_actual):
            await self.testPassed(_description, typecast(_actual, str), typecast(1, str))
        else:
            await self.testFailed(_description, typecast(_actual, str), typecast(1, str))

    async def ensureFalse(self, _description: str, _actual: int) -> None:
        if cond(_actual):
            await self.testFailed(_description, typecast(_actual, str), typecast(0, str))
        else:
            await self.testPassed(_description, typecast(_actual, str), typecast(0, str))

    async def ensureIntegerEqual(self, _description: str, _actual: int, _expected: int) -> None:
        if cond(req(_expected, _actual)):
            await self.testPassed(_description, typecast(_actual, str), typecast(_expected, str))
        else:
            await self.testFailed(_description, typecast(_actual, str), typecast(_expected, str))

    async def floatEqual(self, _actual: float, _expected: float) -> int:
        _error: float = await self.builtin_funcs.llFabs(rsub(_actual, _expected))
        _epsilon: float = 0.001
        if cond(rgreater(_epsilon, _error)):
            print(radd(typecast(_error, str), "Float equality delta "))
            return 0
        return 1

    async def ensureFloatEqual(self, _description: str, _actual: float, _expected: float) -> None:
        if cond(await self.floatEqual(_actual, _expected)):
            await self.testPassed(_description, typecast(_actual, str), typecast(_expected, str))
        else:
            await self.testFailed(_description, typecast(_actual, str), typecast(_expected, str))

    async def ensureStringEqual(self, _description: str, _actual: str, _expected: str) -> None:
        if cond(req(_expected, _actual)):
            await self.testPassed(_description, typecast(_actual, str), typecast(_expected, str))
        else:
            await self.testFailed(_description, typecast(_actual, str), typecast(_expected, str))

    async def ensureVectorEqual(self, _description: str, _actual: Vector, _expected: Vector) -> None:
        if cond(rbooland(await self.floatEqual(_actual[2], _expected[2]), rbooland(await self.floatEqual(_actual[1], _expected[1]), await self.floatEqual(_actual[0], _expected[0])))):
            await self.testPassed(_description, typecast(_actual, str), typecast(_expected, str))
        else:
            await self.testFailed(_description, typecast(_actual, str), typecast(_expected, str))

    async def ensureRotationEqual(self, _description: str, _actual: Quaternion, _expected: Quaternion) -> None:
        if cond(rbooland(await self.floatEqual(_actual[3], _expected[3]), rbooland(await self.floatEqual(_actual[2], _expected[2]), rbooland(await self.floatEqual(_actual[1], _expected[1]), await self.floatEqual(_actual[0], _expected[0]))))):
            await self.testPassed(_description, typecast(_actual, str), typecast(_expected, str))
        else:
            await self.testFailed(_description, typecast(_actual, str), typecast(_expected, str))

    async def ensureListEqual(self, _description: str, _actual: list, _expected: list) -> None:
        if cond(rbooland((req(typecast(_expected, str), typecast(_actual, str))), req(_expected, _actual))):
            await self.testPassed(_description, typecast(_actual, str), typecast(_expected, str))
        else:
            await self.testFailed(_description, typecast(_actual, str), typecast(_expected, str))

    async def callOrderFunc(self, _num: int) -> int:
        self.gCallOrder = radd([_num], self.gCallOrder)
        return 1

    async def testReturn(self) -> int:
        return 1

    async def testReturnFloat(self) -> float:
        return 1.0

    async def testReturnString(self) -> str:
        return "Test string"

    async def testReturnList(self) -> list:
        return [1, 2, 3]

    async def testReturnVector(self) -> Vector:
        return Vector((1.0, 2.0, 3.0))

    async def testReturnRotation(self) -> Quaternion:
        return Quaternion((1.0, 2.0, 3.0, 4.0))

    async def testReturnVectorNested(self) -> Vector:
        return await self.testReturnVector()

    async def testReturnVectorWithLibraryCall(self) -> Vector:
        await self.builtin_funcs.llSin(0.0)
        return Vector((1.0, 2.0, 3.0))

    async def testReturnRotationWithLibraryCall(self) -> Quaternion:
        await self.builtin_funcs.llSin(0.0)
        return Quaternion((1.0, 2.0, 3.0, 4.0))

    async def testParameters(self, _param: int) -> int:
        _param = radd(1, _param)
        return _param

    async def testRecursion(self, _param: int) -> int:
        if cond(rleq(0, _param)):
            return 0
        else:
            return await self.testRecursion(rsub(1, _param))

    async def testExpressionLists(self, _l: list) -> str:
        return radd(typecast(_l, str), "foo")

    @with_goto
    async def tests(self) -> None:
        _a: Optional[list] = None
        _b: Optional[list] = None
        _c: Optional[list] = None
        _i: int = 0
        _v: Vector = Vector((0.0, 0.0, 0.0))
        _q: Quaternion = Quaternion((0.0, 0.0, 0.0, 0.0))
        _l: Optional[list] = None
        _l2: Optional[list] = None
        await self.ensureIntegerEqual("TRUE", 1, 1)
        await self.ensureIntegerEqual("FALSE", 0, 0)
        if cond(0.0):
            await self.testFailed("if(0.0)", "TRUE", "FALSE")
        else:
            await self.testPassed("if(0.0)", "TRUE", "TRUE")
        if cond(0.000001):
            await self.testPassed("if(0.000001)", "TRUE", "TRUE")
        else:
            await self.testFailed("if(0.000001)", "TRUE", "FALSE")
        if cond(0.9):
            await self.testPassed("if(0.9)", "TRUE", "TRUE")
        else:
            await self.testFailed("if(0.9)", "TRUE", "FALSE")
        if cond(Vector((0.0, 0.0, 0.0))):
            await self.testFailed("if(ZERO_VECTOR)", "TRUE", "FALSE")
        else:
            await self.testPassed("if(ZERO_VECTOR)", "TRUE", "TRUE")
        if cond(Quaternion((0.0, 0.0, 0.0, 1.0))):
            await self.testFailed("if(ZERO_ROTATION)", "TRUE", "FALSE")
        else:
            await self.testPassed("if(ZERO_ROTATION)", "TRUE", "TRUE")
        if cond("00000000-0000-0000-0000-000000000000"):
            await self.testPassed("if(NULL_KEY)", "TRUE", "TRUE")
        else:
            await self.testFailed("if(NULL_KEY)", "TRUE", "FALSE")
        if cond(typecast("00000000-0000-0000-0000-000000000000", Key)):
            await self.testFailed("if((key)NULL_KEY)", "TRUE", "FALSE")
        else:
            await self.testPassed("if((key)NULL_KEY)", "TRUE", "TRUE")
        if cond(""):
            await self.testFailed("if(\"\")", "TRUE", "FALSE")
        else:
            await self.testPassed("if(\"\")", "TRUE", "TRUE")
        if cond([]):
            await self.testFailed("if([])", "TRUE", "FALSE")
        else:
            await self.testPassed("if([])", "TRUE", "TRUE")
        await self.ensureIntegerEqual("(TRUE == TRUE)", (req(1, 1)), 1)
        await self.ensureIntegerEqual("(TRUE == FALSE)", (req(0, 1)), 0)
        await self.ensureIntegerEqual("(FALSE == TRUE)", (req(1, 0)), 0)
        await self.ensureIntegerEqual("(FALSE == FALSE)", (req(0, 0)), 1)
        await self.ensureIntegerEqual("(TRUE != TRUE)", (rneq(1, 1)), 0)
        await self.ensureIntegerEqual("(TRUE != FALSE)", (rneq(0, 1)), 1)
        await self.ensureIntegerEqual("(FALSE != TRUE)", (rneq(1, 0)), 1)
        await self.ensureIntegerEqual("(FALSE != FALSE)", (rneq(0, 0)), 0)
        await self.ensureIntegerEqual("(TRUE && TRUE)", (rbooland(1, 1)), 1)
        await self.ensureIntegerEqual("(TRUE && FALSE)", (rbooland(0, 1)), 0)
        await self.ensureIntegerEqual("(FALSE && TRUE)", (rbooland(1, 0)), 0)
        await self.ensureIntegerEqual("(FALSE && FALSE)", (rbooland(0, 0)), 0)
        await self.ensureIntegerEqual("(1 && 2)", (rbooland(2, 1)), 1)
        await self.ensureIntegerEqual("(1 && 0)", (rbooland(0, 1)), 0)
        await self.ensureIntegerEqual("(0 && 2)", (rbooland(2, 0)), 0)
        await self.ensureIntegerEqual("(0 && 0)", (rbooland(0, 0)), 0)
        await self.ensureIntegerEqual("(TRUE || TRUE)", (rboolor(1, 1)), 1)
        await self.ensureIntegerEqual("(TRUE || FALSE)", (rboolor(0, 1)), 1)
        await self.ensureIntegerEqual("(FALSE || TRUE)", (rboolor(1, 0)), 1)
        await self.ensureIntegerEqual("(FALSE || FALSE)", (rboolor(0, 0)), 0)
        await self.ensureIntegerEqual("(1 || 2)", (rboolor(2, 1)), 1)
        await self.ensureIntegerEqual("(1 || 0)", (rboolor(0, 1)), 1)
        await self.ensureIntegerEqual("(0 || 2)", (rboolor(2, 0)), 1)
        await self.ensureIntegerEqual("(0 || 0)", (rboolor(0, 0)), 0)
        await self.ensureIntegerEqual("(! TRUE)", (boolnot(1)), 0)
        await self.ensureIntegerEqual("(! FALSE)", (boolnot(0)), 1)
        await self.ensureIntegerEqual("(! 2)", (boolnot(2)), 0)
        await self.ensureIntegerEqual("(! 0)", (boolnot(0)), 1)
        await self.ensureIntegerEqual("(1 > 0)", (rgreater(0, 1)), 1)
        await self.ensureIntegerEqual("(0 > 1)", (rgreater(1, 0)), 0)
        await self.ensureIntegerEqual("(1 > 1)", (rgreater(1, 1)), 0)
        await self.ensureIntegerEqual("(0 < 1)", (rless(1, 0)), 1)
        await self.ensureIntegerEqual("(1 < 0)", (rless(0, 1)), 0)
        await self.ensureIntegerEqual("(1 < 1)", (rless(1, 1)), 0)
        await self.ensureIntegerEqual("(1 >= 0)", (rgeq(0, 1)), 1)
        await self.ensureIntegerEqual("(0 >= 1)", (rgeq(1, 0)), 0)
        await self.ensureIntegerEqual("(1 >= 1)", (rgeq(1, 1)), 1)
        await self.ensureIntegerEqual("(0 <= 1)", (rleq(1, 0)), 1)
        await self.ensureIntegerEqual("(1 <= 0)", (rleq(0, 1)), 0)
        await self.ensureIntegerEqual("(1 <= 1)", (rleq(1, 1)), 1)
        await self.ensureIntegerEqual("(10 & 25)", (rbitand(25, 10)), 8)
        await self.ensureIntegerEqual("(10 | 25)", (rbitor(25, 10)), 27)
        await self.ensureIntegerEqual("~10", bitnot(10), -11)
        await self.ensureIntegerEqual("(10 ^ 25)", (rbitxor(25, 10)), 19)
        await self.ensureIntegerEqual("(523 >> 2)", (rshr(2, 523)), 130)
        await self.ensureIntegerEqual("(523 << 2)", (rshl(2, 523)), 2092)
        await self.ensureIntegerEqual("(1 + 1)", (radd(1, 1)), 2)
        await self.ensureFloatEqual("(1 + 1.1)", (radd(1.1, 1.0)), 2.1)
        await self.ensureFloatEqual("(1.1 + 1)", (radd(1.0, 1.1)), 2.1)
        await self.ensureFloatEqual("(1.1 + 1.1)", (radd(1.1, 1.1)), 2.2)
        await self.ensureStringEqual("\"foo\" + \"bar\"", radd("bar", "foo"), "foobar")
        await self.ensureVectorEqual("(<1.1, 2.2, 3.3> + <4.4, 5.5, 6.6>)", (radd(Vector((4.4, 5.5, 6.6)), Vector((1.1, 2.2, 3.3)))), Vector((5.5, 7.7, 9.9)))
        await self.ensureRotationEqual("(<1.1, 2.2, 3.3, 4.4> + <4.4, 5.5, 6.6, 3.3>)", (radd(Quaternion((4.4, 5.5, 6.6, 3.3)), Quaternion((1.1, 2.2, 3.3, 4.4)))), Quaternion((5.5, 7.7, 9.9, 7.7)))
        await self.ensureListEqual("([1] + 2)", (radd(2, [1])), [1, 2])
        await self.ensureListEqual("([] + 1.5)", (radd(1.5, [])), [1.5])
        await self.ensureListEqual("([\"foo\"] + \"bar\")", (radd("bar", ["foo"])), ["foo", "bar"])
        await self.ensureListEqual("([] + <1,2,3>)", (radd(Vector((1.0, 2.0, 3.0)), [])), [Vector((1.0, 2.0, 3.0))])
        await self.ensureListEqual("([] + <1,2,3,4>)", (radd(Quaternion((1.0, 2.0, 3.0, 4.0)), [])), [Quaternion((1.0, 2.0, 3.0, 4.0))])
        await self.ensureListEqual("(1 + [2])", (radd([2], 1)), [1, 2])
        await self.ensureListEqual("(1.0 + [2])", (radd([2], 1.0)), [1.0, 2])
        await self.ensureListEqual("(1 + [2])", (radd([2], "one")), ["one", 2])
        await self.ensureListEqual("(<1.0,1.0,1.0,1.0> + [2])", (radd([2], Quaternion((1.0, 1.0, 1.0, 1.0)))), [Quaternion((1.0, 1.0, 1.0, 1.0)), 2])
        await self.ensureListEqual("(<1.0,1.0,1.0> + [2])", (radd([2], Vector((1.0, 1.0, 1.0)))), [Vector((1.0, 1.0, 1.0)), 2])
        _a = []
        _b = _a
        _a = radd(["foo"], _a)
        await self.ensureListEqual("list a = []; list b = a; a += [\"foo\"]; a == [\"foo\"]", _a, ["foo"])
        await self.ensureListEqual("list a = []; list b = a; a += [\"foo\"]; b == []", _b, [])
        _a = ["a"]
        _b = ["b"]
        _c = radd(_b, _a)
        await self.ensureListEqual("a = [\"a\"]; b = [\"b\"]; list c = a + b; a == [\"a\"];", _a, ["a"])
        await self.ensureListEqual("a = [\"a\"]; b = [\"b\"]; list c = a + b; b == [\"b\"];", _b, ["b"])
        await self.ensureListEqual("a = [\"a\"]; b = [\"b\"]; list c = a + b; c == [\"a\", \"b\"];", _c, ["a", "b"])
        await self.ensureIntegerEqual("(1 - 1)", (rsub(1, 1)), 0)
        await self.ensureFloatEqual("(1 - 0.5)", (rsub(0.5, 1.0)), 0.5)
        await self.ensureFloatEqual("(1.5 - 1)", (rsub(1.0, 1.5)), 0.5)
        await self.ensureFloatEqual("(2.2 - 1.1)", (rsub(1.1, 2.2)), 1.1)
        await self.ensureVectorEqual("(<1.5, 2.5, 3.5> - <4.5, 5.5, 6.5>)", (rsub(Vector((4.5, 5.5, 6.5)), Vector((1.5, 2.5, 3.5)))), Vector((-3.0, -3.0, -3.0)))
        await self.ensureRotationEqual("(<1.5, 2.5, 3.5, 4.5> - <4.5, 5.5, 6.5, 7.5>)", (rsub(Quaternion((4.5, 5.5, 6.5, 7.5)), Quaternion((1.5, 2.5, 3.5, 4.5)))), Quaternion((-3.0, -3.0, -3.0, -3.0)))
        await self.ensureIntegerEqual("(2 * 3)", (rmul(3, 2)), 6)
        await self.ensureFloatEqual("(2 * 3.5)", (rmul(3.5, 2.0)), 7.0)
        await self.ensureFloatEqual("(2.5 * 3)", (rmul(3.0, 2.5)), 7.5)
        await self.ensureFloatEqual("(2.5 * 3.5)", (rmul(3.5, 2.5)), 8.75)
        await self.ensureVectorEqual("(<1.1, 2.2, 3.3> * 2)", (rmul(2.0, Vector((1.1, 2.2, 3.3)))), Vector((2.2, 4.4, 6.6)))
        await self.ensureVectorEqual("(2 * <1.1, 2.2, 3.3>)", (rmul(2.0, Vector((1.1, 2.2, 3.3)))), (rmul(Vector((1.1, 2.2, 3.3)), 2.0)))
        await self.ensureVectorEqual("(<2.2, 4.4, 6.6> * 2.0)", (rmul(2.0, Vector((2.2, 4.4, 6.6)))), Vector((4.4, 8.8, 13.2)))
        await self.ensureVectorEqual("(2.0 * <2.2, 4.4, 6.6>)", (rmul(2.0, Vector((2.2, 4.4, 6.6)))), (rmul(Vector((2.2, 4.4, 6.6)), 2.0)))
        await self.ensureFloatEqual("<1,3,-5> * <4,-2,-1>", rmul(Vector((4.0, -2.0, -1.0)), Vector((1.0, 3.0, -5.0))), 3.0)
        await self.ensureVectorEqual("<-1,0,0> * <0, 0, 0.707, 0.707>", rmul(Quaternion((0.0, 0.0, 0.707, 0.707)), Vector((-1.0, 0.0, 0.0))), Vector((0.0, -1.0, 0.0)))
        await self.ensureRotationEqual("(<1.0, 2.0, 3.0, 4.0> * <5.0, 6.0, 7.0, 8.0>)", (rmul(Quaternion((5.0, 6.0, 7.0, 8.0)), Quaternion((1.0, 2.0, 3.0, 4.0)))), Quaternion((32.0, 32.0, 56.0, -6.0)))
        await self.ensureIntegerEqual("(2 / 2)", (rdiv(2, 2)), 1)
        await self.ensureFloatEqual("(2.2 / 2)", (rdiv(2.0, 2.2)), 1.1)
        await self.ensureFloatEqual("(3 / 1.5)", (rdiv(1.5, 3.0)), 2.0)
        await self.ensureFloatEqual("(2.2 / 2.0)", (rdiv(2.0, 2.2)), 1.1)
        await self.ensureVectorEqual("(<1.0, 2.0, 3.0> / 2)", (rdiv(2.0, Vector((1.0, 2.0, 3.0)))), Vector((0.5, 1.0, 1.5)))
        await self.ensureVectorEqual("(<3.0, 6.0, 9.0> / 1.5)", (rdiv(1.5, Vector((3.0, 6.0, 9.0)))), Vector((2.0, 4.0, 6.0)))
        await self.ensureVectorEqual("<-1,0,0> / <0, 0, 0.707, 0.707>", rdiv(Quaternion((0.0, 0.0, 0.707, 0.707)), Vector((-1.0, 0.0, 0.0))), Vector((0.0, 1.0, 0.0)))
        await self.ensureRotationEqual("(<1.0, 2.0, 3.0, 4.0> / <5.0, 6.0, 7.0, 8.0>)", (rdiv(Quaternion((5.0, 6.0, 7.0, 8.0)), Quaternion((1.0, 2.0, 3.0, 4.0)))), Quaternion((-16.0, 0.0, -8.0, 70.0)))
        await self.ensureIntegerEqual("(3 % 1)", (rmod(1, 3)), 0)
        await self.ensureVectorEqual("(<1.0, 2.0, 3.0> % <4.0, 5.0, 6.0>)", (rmod(Vector((4.0, 5.0, 6.0)), Vector((1.0, 2.0, 3.0)))), Vector((-3.0, 6.0, -3.0)))
        _i = 1
        await self.ensureIntegerEqual("i = 1;", _i, 1)
        _i = 1
        _i = radd(1, _i)
        await self.ensureIntegerEqual("i = 1; i += 1;", _i, 2)
        _i = 1
        _i = rsub(1, _i)
        await self.ensureIntegerEqual("i = 1; i -= 1;", _i, 0)
        _i = 2
        _i = rmul(2, _i)
        await self.ensureIntegerEqual("i = 2; i *= 2;", _i, 4)
        _i = 1
        (_i := typecast(rmul(0.5, _i), int))
        await self.ensureIntegerEqual("i = 1; i *= 0.5;", _i, 0)
        _i = 2
        _i = rdiv(2, _i)
        await self.ensureIntegerEqual("i = 2; i /= 2;", _i, 1)
        _i = rdiv(-1, -2147483648)
        await self.ensureIntegerEqual("i = 0x80000000 / -1;", _i, -2147483648)
        _i = 3
        _i = rmod(1, _i)
        await self.ensureIntegerEqual("i = 3; i %= 1;", _i, 0)
        _i = rmod(-1, -2147483648)
        await self.ensureIntegerEqual("i = 0x80000000 % -1;", _i, 0)
        _i = 1
        await self.ensureIntegerEqual("postinc", rbooland((req(1, postincr(locals(), "_i"))), (req(2, _i))), 1)
        _i = 1
        await self.ensureIntegerEqual("preinc", rbooland((req(2, preincr(locals(), "_i"))), (req(2, _i))), 1)
        _i = 2
        await self.ensureIntegerEqual("postdec", rbooland((req(2, postdecr(locals(), "_i"))), (req(1, _i))), 1)
        _i = 2
        await self.ensureIntegerEqual("predec1", rbooland((req(1, predecr(locals(), "_i"))), (req(1, _i))), 1)
        _i = 2
        _i -= 1
        await self.ensureIntegerEqual("predec2", _i, 1)
        await self.ensureFloatEqual("((float)2)", (2.0), 2.0)
        await self.ensureStringEqual("((string)2)", (typecast(2, str)), "2")
        await self.ensureIntegerEqual("((integer) 1.5)", (typecast(1.5, int)), 1)
        await self.ensureStringEqual("((string) 1.5)", (typecast(1.5, str)), "1.500000")
        await self.ensureIntegerEqual("((integer) \"0xF\")", (typecast("0xF", int)), 15)
        await self.ensureIntegerEqual("((integer) \"2\")", (typecast("2", int)), 2)
        await self.ensureFloatEqual("((float) \"1.5\")", (typecast("1.5", float)), 1.5)
        await self.ensureVectorEqual("((vector) \"<1,2,3>\")", (typecast("<1,2,3>", Vector)), Vector((1.0, 2.0, 3.0)))
        await self.ensureRotationEqual("((quaternion) \"<1,2,3,4>\")", (typecast("<1,2,3,4>", Quaternion)), Quaternion((1.0, 2.0, 3.0, 4.0)))
        await self.ensureStringEqual("((string) <1,2,3>)", (typecast(Vector((1.0, 2.0, 3.0)), str)), "<1.00000, 2.00000, 3.00000>")
        await self.ensureStringEqual("((string) <1,2,3,4>)", (typecast(Quaternion((1.0, 2.0, 3.0, 4.0)), str)), "<1.00000, 2.00000, 3.00000, 4.00000>")
        await self.ensureStringEqual("((string) [1,2.5,<1,2,3>])", (typecast([1, 2.5, Vector((1.0, 2.0, 3.0))], str)), "12.500000<1.000000, 2.000000, 3.000000>")
        _i = 0
        while cond(rless(10, _i)):
            _i += 1
        await self.ensureIntegerEqual("i = 0; while(i < 10) ++i", _i, 10)
        _i = 0
        while True == True:
            _i += 1
            if not cond(rless(10, _i)):
                break
        await self.ensureIntegerEqual("i = 0; do {++i;} while(i < 10);", _i, 10)
        _i = 0
        while True == True:
            if not cond(rless(10, _i)):
                break
            pass
            _i += 1
        await self.ensureIntegerEqual("for(i = 0; i < 10; ++i);", _i, 10)
        _i = 1
        goto ._SkipAssign
        await self.builtin_funcs.llSetText("Error", Vector((1.0, 0.0, 0.0)), 1.0)
        _i = 2
        label ._SkipAssign
        await self.ensureIntegerEqual("assignjump", _i, 1)
        await self.ensureIntegerEqual("testReturn()", await self.testReturn(), 1)
        await self.ensureFloatEqual("testReturnFloat()", await self.testReturnFloat(), 1.0)
        await self.ensureStringEqual("testReturnString()", await self.testReturnString(), "Test string")
        await self.ensureVectorEqual("testReturnVector()", await self.testReturnVector(), Vector((1.0, 2.0, 3.0)))
        await self.ensureRotationEqual("testReturnRotation()", await self.testReturnRotation(), Quaternion((1.0, 2.0, 3.0, 4.0)))
        await self.ensureListEqual("testReturnList()", await self.testReturnList(), [1, 2, 3])
        await self.ensureVectorEqual("testReturnVectorNested()", await self.testReturnVectorNested(), Vector((1.0, 2.0, 3.0)))
        await self.ensureVectorEqual("libveccall", await self.testReturnVectorWithLibraryCall(), Vector((1.0, 2.0, 3.0)))
        await self.ensureRotationEqual("librotcall", await self.testReturnRotationWithLibraryCall(), Quaternion((1.0, 2.0, 3.0, 4.0)))
        await self.ensureIntegerEqual("testParameters(1)", await self.testParameters(1), 2)
        _i = 1
        await self.ensureIntegerEqual("i = 1; testParameters(i)", await self.testParameters(_i), 2)
        await self.ensureIntegerEqual("testRecursion(10)", await self.testRecursion(10), 0)
        await self.ensureIntegerEqual("gInteger", self.gInteger, 5)
        await self.ensureFloatEqual("gFloat", self.gFloat, 1.5)
        await self.ensureStringEqual("gString", self.gString, "foo")
        await self.ensureVectorEqual("gVector", self.gVector, Vector((1.0, 2.0, 3.0)))
        await self.ensureRotationEqual("gRot", self.gRot, Quaternion((1.0, 2.0, 3.0, 4.0)))
        await self.ensureListEqual("gList", self.gList, [1, 2, 3])
        self.gInteger = 1
        await self.ensureIntegerEqual("gInteger = 1", self.gInteger, 1)
        self.gFloat = 0.5
        await self.ensureFloatEqual("gFloat = 0.5", self.gFloat, 0.5)
        self.gString = "bar"
        await self.ensureStringEqual("gString = \"bar\"", self.gString, "bar")
        self.gVector = Vector((3.0, 3.0, 3.0))
        await self.ensureVectorEqual("gVector = <3,3,3>", self.gVector, Vector((3.0, 3.0, 3.0)))
        self.gRot = Quaternion((3.0, 3.0, 3.0, 3.0))
        await self.ensureRotationEqual("gRot = <3,3,3,3>", self.gRot, Quaternion((3.0, 3.0, 3.0, 3.0)))
        self.gList = [4, 5, 6]
        await self.ensureListEqual("gList = [4,5,6]", self.gList, [4, 5, 6])
        self.gVector = Vector((3.0, 3.0, 3.0))
        await self.ensureVectorEqual("-gVector = <-3,-3,-3>", neg(self.gVector), Vector((-3.0, -3.0, -3.0)))
        self.gRot = Quaternion((3.0, 3.0, 3.0, 3.0))
        await self.ensureRotationEqual("-gRot = <-3,-3,-3,-3>", neg(self.gRot), Quaternion((-3.0, -3.0, -3.0, -3.0)))
        _v = Vector((0.0, 0.0, 0.0))
        _v = replace_coord_axis(_v, 0, 3.0)
        await self.ensureFloatEqual("v.x", _v[0], 3.0)
        _q = Quaternion((0.0, 0.0, 0.0, 1.0))
        _q = replace_coord_axis(_q, 3, 5.0)
        await self.ensureFloatEqual("q.s", _q[3], 5.0)
        self.gVector = replace_coord_axis(self.gVector, 1, 17.5)
        await self.ensureFloatEqual("gVector.y = 17.5", self.gVector[1], 17.5)
        self.gRot = replace_coord_axis(self.gRot, 2, 19.5)
        await self.ensureFloatEqual("gRot.z = 19.5", self.gRot[2], 19.5)
        _l = typecast(5, list)
        _l2 = typecast(5, list)
        await self.ensureListEqual("leq1", _l, _l2)
        await self.ensureListEqual("leq2", _l, [5])
        await self.ensureListEqual("leq3", [1.5, 6, Vector((1.0, 2.0, 3.0)), Quaternion((1.0, 2.0, 3.0, 4.0))], [1.5, 6, Vector((1.0, 2.0, 3.0)), Quaternion((1.0, 2.0, 3.0, 4.0))])
        await self.ensureIntegerEqual("sesc1", await self.builtin_funcs.llStringLength("\\"), 1)
        await self.ensureIntegerEqual("sesc2", await self.builtin_funcs.llStringLength("    "), 4)
        await self.ensureIntegerEqual("sesc3", await self.builtin_funcs.llStringLength("\n"), 1)
        await self.ensureIntegerEqual("sesc4", await self.builtin_funcs.llStringLength("\""), 1)
        await self.ensureStringEqual("testExpressionLists([testExpressionLists([]), \"bar\"]) == \"foofoobar\"", await self.testExpressionLists([await self.testExpressionLists([]), "bar"]), "foofoobar")
        if cond(rboolor(1, rbooland(rbitor(rbitxor(rdiv(await self.callOrderFunc(5), await self.callOrderFunc(4)), await self.callOrderFunc(3)), await self.callOrderFunc(2)), rboolor(rmul(await self.callOrderFunc(1), await self.callOrderFunc(0)), 1)))):
            pass
        await self.ensureListEqual("gCallOrder expected order", self.gCallOrder, [5, 4, 3, 2, 1, 0])
        await self.ensureIntegerEqual("(gInteger = 5)", (assign(self.__dict__, "gInteger", 5)), 5)
        await self.ensureFloatEqual("(gVector.z = 6)", (assign(self.__dict__, "gVector", replace_coord_axis(self.gVector, 2, 6.0))[2]), 6.0)
        self.gVector = Vector((1.0, 2.0, 3.0))
        await self.ensureFloatEqual("++gVector.z", preincr(self.__dict__, "gVector", 2), 4.0)
        self.gVector = Vector((1.0, 2.0, 3.0))
        await self.ensureFloatEqual("gVector.z++", postincr(self.__dict__, "gVector", 2), 3.0)
        await self.ensureFloatEqual("(v.z = 6)", ((_v := replace_coord_axis(_v, 2, 6.0))[2]), 6.0)
        _v = Vector((1.0, 2.0, 3.0))
        await self.ensureFloatEqual("++v.z", preincr(locals(), "_v", 2), 4.0)
        _v = Vector((1.0, 2.0, 3.0))
        await self.ensureFloatEqual("v.z++", postincr(locals(), "_v", 2), 3.0)
        await self.ensureFloatEqual("posinf == posinf", float("inf"), float("inf"))
        await self.ensureFloatEqual("neginf == neginf", float("-inf"), float("-inf"))
        await self.ensureFalse("posinf != neginf", req(float("-inf"), float("inf")))
        await self.ensureStringEqual("(string)posinf == 'Infinity'", typecast(float("inf"), str), "Infinity")
        await self.ensureStringEqual("(string)neginf == '-Infinity'", typecast(float("-inf"), str), "-Infinity")

    async def runTests(self) -> None:
        self.gInteger = 5
        self.gFloat = 1.5
        self.gString = "foo"
        self.gVector = Vector((1.0, 2.0, 3.0))
        self.gRot = Quaternion((1.0, 2.0, 3.0, 4.0))
        self.gList = [1, 2, 3]
        self.gTestsPassed = 0
        self.gTestsFailed = 0
        self.gCallOrder = []
        await self.tests()
        print("All tests passed")

    async def edefaultstate_entry(self) -> None:
        await self.runTests()




#    (C) Copyright 2015-2022 Sei Lisa. All rights reserved.
#
#    This file is part of LSL PyOptimizer.
#
#    LSL PyOptimizer is free software: you can redistribute it and/or
#    modify it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    LSL PyOptimizer is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with LSL PyOptimizer. If not, see <http://www.gnu.org/licenses/>.

# JSON functions

import re
import math
from .lslcommon import *
from .lslbasefuncs import llStringTrim, fs, fl, InternalTypecast

# INCOMPATIBILITY NOTE: The JSON functions in SL have very weird behaviour
# in corner cases. Despite our best efforts, that behaviour is not replicated
# here, as doing so proved to be too difficult to investigate and implement.
# The functions in here behave somewhat more sanely in these corner cases than
# in SL, and may therefore fail to reproduce the same results as SL does.

# If you wish to maintain compatibility, you can disable the JSON functions
# by commenting out the 'from lsljson import *' line in lslfuncs.py.

JSON_INVALID = u'\uFDD0'
JSON_OBJECT  = u'\uFDD1'
JSON_ARRAY   = u'\uFDD2'
JSON_NUMBER  = u'\uFDD3'
JSON_STRING  = u'\uFDD4'
JSON_NULL    = u'\uFDD5'
JSON_TRUE    = u'\uFDD6'
JSON_FALSE   = u'\uFDD7'
JSON_DELETE  = u'\uFDD8'

JSON_APPEND  = -1

jsonesc_re = re.compile(u'[\x08\x09\x0A\x0C\x0D"/\\\\]')
jsonesc_dict = {u'\x08':u'\\b', u'\x09':u'\\t', u'\x0A':u'\\n', u'\x0C':u'\\f',
                u'\x0D':u'\\r', u'"':u'\\"', u'/':u'\\/', u'\\':u'\\\\'}
jsonunesc_dict = {u'b':u'\x08', u't':u'\x09', u'n':u'\x0A', u'f':u'\x0C', u'r':u'\x0D'}

# LSL JSON numbers differ from standard JSON numbers in many respects:
#      Numbers starting with 0 are allowed, e.g. 01.3e4, 00042
#      .5 is allowed.
#      1e+0 is NOT allowed (the + after the e, to be precise). BUG-6466.
#      . is allowed, as is -.e-0 etc.
#      1E is allowed.
#      E.2 is allowed.
#      E is allowed.
#      1E-1.2 is allowed.
# In general, the rule seems to be: at most one 'E' (optionally followed by a
# '-') and one '.', with optional digits interspersed and an optional initial
# minus sign.
#
# Our RE below checks for the two possible orders of '.' and 'E'. One branch
# must have a mandatory 'E'; in the other everything is optional but it must
# have at least 1 character (done by the lookahead assertion).
#
# The capturing groups serve to check whether the first variant was taken, and
# whether there is something after the digits in the second variant. If both
# are empty, then the match is just one or more digits preceded by an optional
# minus sign (i.e. an integer). That's used by llJson2List to return integer
# elements when appropriate.

# Real JSON number parser:
#jsonnum_re = re.compile(str2u(
#    r'-?(?:[1-9][0-9]*|0)(?:\.[0-9]+)?(?:[Ee][+-]?[0-9]+)?'
#    ))

# BUG-6466 active:
jsonnumbug_re = re.compile(str2u(r'''
    -?(?:
        [0-9]*([Ee])-?[0-9]*\.?[0-9]*
        |(?=[0-9Ee.])[0-9]*(\.?[0-9]*(?:[Ee]-?)?[0-9]*)
    )
    '''), re.X)
# BUG-6466 fixed:
# The new RE is just a modified version of the crap, allowing + exponents and
# disallowing zeros, sometimes even when legal (e.g. 0e0)
#jsonnum_re = re.compile(str2u(r'''
#    -?(?:
#        (?=[1-9]|\.(?:[^e]|$)
#            |0(?:[^0-9e]|$))[0-9]*([Ee])[+-]?[0-9]*\.?[0-9]*
#        |(?=[1-9]|\.(?:[^e]|$)
#            |0(?:[^0-9e]|$))[0-9]*(\.?[0-9]*(?:[Ee][+-]?)?[0-9]*)
#    )
#    '''), re.X)
# They've fixed BUG-6657 by bringing BUG-6466 back to life.
jsonnum_re = re.compile(str2u(r'''
    -?(?:
        [0-9]*([Ee])-?[0-9]*\.?[0-9]*
        |(?=[0-9Ee.])[0-9]*(\.?[0-9]*(?:[Ee]-?)?[0-9]*)
    )
    '''), re.X)

jsonstring_re = re.compile(str2u(r'"(?:[^"\\]|\\.)*"'))

# This might need some explanation. The ] and - are included in the first
# set, the ] in the first after the ^ and the - in the last positions of
# the set as required by RE syntax. The [ is part of it and isn't special,
# though it confuses things. The set comprises any character not in
# -{}[],:"0123456789
# The second set comprises zero or more characters not in ,:]}
#word_re = re.compile(str2u(r'[^][{}0-9",:-][^]},:]*'))
# Screw that, we're using just a fallback.
jsoncatchall_re = re.compile(u'(.*?)[\x09\x0A\x0B\x0C\x0D ]*(?:[]},]|$)')

digits_re = re.compile(u'[0-9]{1,9}')


class EInternalJsonInvalid(Exception):
    """Used to force return of JSON_INVALID from child functions"""
    pass

def InternalJsonQuote(s):
    return u'"' + jsonesc_re.sub(lambda x: jsonesc_dict[x.group()], s) + u'"'

def InternalJsonUnquote(s):
    """Relaxed unquote with LSL rules. Assumes string starts and ends in ",
    may contain " and may end in \" too (i.e. malformed). E.g. "a"b\" is a
    valid string for this function and the result is a"b\
    """
    assert s != u''
    assert s[0] == s[-1] == u'"' and s[1:2]

    ret = u''
    esc = False
    for c in s[1:-1]:
        if esc:
            try:
                ret += jsonunesc_dict[c]
            except KeyError:
                ret += c
            esc = False
        else:
            if c == u'\\':
                esc = True
            else:
                ret += c
    if esc:
        return ret + u'\\'
    return ret

def InternalJsonUnquoteX(s):
    """Rigorous unquote; checks for quotes at the beginning and end only."""
    esc = last = False
    first = True

    ret = u''
    for c in s:
        if last:
            break
        if esc:
            try:
                ret += jsonunesc_dict[c]
            except:
                ret += c
            esc = False
            first = False
        elif first:
            if c != u'"': break
            first = False
        elif c == u'"':
            last = True
            first = False
        elif c == u'\\':
            esc = True
        else:
            ret += c
    else:
        if not first and last:
            return ret
    return s # malformed string, return the original

def InternalJsonF2S(f):
    if math.isnan(f):
        return u'nan'
    if math.isinf(f):
        return u'inf' if f > 0 else u'-inf'
    return u'%.6f' % f

def InternalJsonScanMatching(json, idx):
    """Shortcut: scan for a matching pair of {} or [] with proper nesting
    and string handling, with no validity check other than well-formedness,
    meaning all {} or [] must match.
    """
    matching = json[idx]
    matching += '}' if json[idx] == '{' else ']'
    level = 1
    str = False
    esc = False
    for i in xrange(idx+1, len(json)):
        c = json[i]
        if str:
            if esc:
                esc = False
            elif c == u'\\':
                esc = True
            elif c == u'"':
                str = False
        elif c == u'"':
            str = True
        elif c in matching:
            if c == matching[0]:
                level += 1
            else:
                level -= 1
            if not level:
                return i+1
    return None

def InternalElement2Json(elem, ParseNumbers = True):
    telem = type(elem)
    if telem == unicode:
        elem = llStringTrim(elem, 3) # STRING_TRIM
        if elem == u'':
            return u'""'
        # Yes, these are checked after trimming. Don't facepalm too hard.
        if elem == JSON_NULL:
            return u'null'
        if elem == JSON_TRUE:
            return u'true'
        if elem == JSON_FALSE:
            return u'false'
        if elem[0] == elem[-1] == u'"' and elem[1:2] or elem in ('null','false','true') \
                or elem[0] == u'[' and elem[-1] == u']' \
                or elem[0] == u'{' and elem[-1] == u'}':
            return elem

        if ParseNumbers:
            match = (jsonnumbug_re if 6466 in Bugs else jsonnum_re).match(elem)
            if match and match.end() == len(elem):
                return elem

        if elem == JSON_INVALID:
            return u''

        return InternalJsonQuote(elem)

    if telem == Key:
        return u'"' + unicode(elem) + u'"'
    if telem in (Vector, Quaternion):
        return u'"<' + u', '.join([InternalJsonF2S(x) for x in elem]) + u'>"'
    if telem == float:
        return InternalJsonF2S(elem)
    # Integer
    return unicode(elem)

def InternalJsonGetToken(json, idx):

    #start = idx
    num_re = jsonnumbug_re if 6466 in Bugs else jsonnum_re

    L = len(json)
    while idx < L:
        c = json[idx]
        if c not in u'\x09\x0A\x0B\x0C\x0D ':
            break
        idx += 1

    if idx >= L:
        return (idx, idx, None)

    c = json[idx]
    if c in u',:{}[]':
        return (idx, idx+1, c)

    match = jsonstring_re.match(json, idx)
    if match:
        return (idx, match.end(), JSON_STRING)

    match = num_re.match(json, idx)
    if match:
        return (idx, match.end(), JSON_NUMBER)

    match = jsoncatchall_re.match(json, idx) # matches always, even if empty string
    s = match.group(1)
    if s in (u'null', u'true', u'false'):
        return (idx, match.end(1),
            JSON_NULL if s == u'null' else JSON_TRUE if s == u'true' else JSON_FALSE)
    return (idx, match.end(1), JSON_INVALID)

def InternalJsonGetTokenFull(json, idx):
    ret = InternalJsonGetToken(json, idx)
    if ret[2] in (u'{', u'['):
        match = InternalJsonScanMatching(json, ret[0])
        if match is not None:
            return (ret[0], match, JSON_OBJECT if ret[2] == u'{' else JSON_ARRAY)
    return ret

def InternalJsonPathMatches(key, pathelem):
    if type(key) == type(pathelem) == int or type(key) == unicode and isinstance(pathelem, unicode):
        return key == pathelem
    if type(key) == unicode and type(pathelem) == int:
        raise EInternalJsonInvalid
    # one combo remains - key is numeric and pathelem is unicode or Key
    match = digits_re.match(pathelem)
    if not match:
        raise EInternalJsonInvalid
    return key == int(match.group())

def InternalJsonFindValue(json, tgtpath, ReturnsToken, SetRules = False):

    # Building a function that meets the strange requisites of LL's json is not easy.
    # These requisites include syntax-checking of all items at the current level,
    # but not of items at a deeper nesting level.

    # Making it one-pass iterative O(len) instead of recursive O(depth*len) is even
    # more of a challenge, especially with these constraints.

    token = InternalJsonGetToken(json, 0)

    if tgtpath == []:
        # No nesting receives special treatment.
        if token[2] in (JSON_NUMBER, JSON_STRING, JSON_NULL, JSON_TRUE, JSON_FALSE, JSON_INVALID):
            if InternalJsonGetToken(json, token[1])[2] is None:
                if ReturnsToken:
                    return token
                if token[2] == JSON_NUMBER:
                    return json[token[0]:token[1]]
                if token[2] == JSON_STRING:
                    return InternalJsonUnquote(json[token[0]:token[1]])
                if token[2] == JSON_INVALID:
                    # Accept malformed strings if they start and end in quotes
                    s = json[token[0]:token[1]]
                    if s[1:2] and s[0] == s[-1] == u'"':
                        return InternalJsonUnquote(s)
                return token[2]
            return JSON_INVALID
        if token[2] not in (u'{', u'['):
            return JSON_INVALID

        json = llStringTrim(json, 2) # STRING_TRIM_RIGHT
        if json[-1] == u'}' and token[2] == u'{':
            if ReturnsToken:
                return (token[0], len(json), JSON_OBJECT)
            return json[token[0]:]
        if json[-1] == u']' and token[2] == u'[':
            if ReturnsToken:
                return (token[0], len(json), JSON_ARRAY)
            return json[token[0]:]
        return JSON_INVALID

        # This would be the code if there was proper scanning.
        #match = InternalJsonScanMatching(json, token[0])
        #if match is None or InternalJsonGetToken(json, match)[2] is not None:
        #    return JSON_INVALID
        #if ReturnsType: # this has been changed tho' - review if ever used
        #    return JSON_OBJECT if token[2] == u'{' else JSON_ARRAY
        #return json[token[0]:match]

    if token[2] not in (u'{', u'['):
        return JSON_INVALID

    # Follow the path
    L = len(tgtpath)
    # For the current position, matchlvl keeps track of how many levels are
    # matched. When matchlvl == L, we are at the item of interest.
    # For example: if we're at the ! in [1.0, "y", true, [1, ![6], {"a":5}]]
    # and the path is [3, 2, "a"], matchlvl will be 1 (meaning the first level
    # of the path, i.e. position 3, is matched, but we're not in sub-position
    # 2 yet).
    matchlvl = 0
    ret = None # the target token, if found, or None if not

    # Keeps track of what we have opened so far.
    stk = [token[2]]

    # This tracks the current key within an array or object. Here we assume
    # it's an array; if it's an object, the item key will replace it anyway.
    curkey = 0

    just_open = True
    just_closed = False

    # Load next token
    token = InternalJsonGetToken(json, token[1])

    try:
        while True:
            # Process value if it can be present
            kind = token[2]
            if not (just_closed or
                    just_open and kind in (u'}', u']')):
                # Item processing.
                # Not entering here immediately after a } or ] (just_closed)
                # or after a { or [ followed by } or ] (just_open...)
                just_open = False
                if kind in u':,]}' or kind == JSON_INVALID:
                    return JSON_INVALID
                if stk[-1] == u'{':
                    # Read the current key
                    if kind != JSON_STRING:
                        return JSON_INVALID
                    colon = InternalJsonGetToken(json, token[1])
                    if colon[2] != u':':
                        return JSON_INVALID
                    curkey = InternalJsonUnquote(json[token[0]:token[1]])
                    token = InternalJsonGetToken(json, colon[1])
                    kind = token[2]
                    del colon
                if matchlvl < L and InternalJsonPathMatches(curkey, tgtpath[matchlvl]):
                    # Descend to this level
                    matchlvl += 1
                    ret = None # because e.g. llJsonGetValue("{\"a\":[1],\"a\":2}",["a",0])==JSON_INVALID
                    if matchlvl == L:
                        if kind in u'{[':
                            match = InternalJsonScanMatching(json, token[0])
                            if match is None:
                                return JSON_INVALID
                            token = (token[0], match, JSON_OBJECT if token[2] == u'{' else JSON_ARRAY)
                        ret = token
                        matchlvl -= 1
                    elif kind in u'{[':
                        stk.append(token[2])
                        curkey = 0
                        just_open = True
                        token = InternalJsonGetToken(json, token[1])
                        continue
                else:
                    # We're skipping the element
                    if kind in u'[{':
                        match = InternalJsonScanMatching(json, token[0])
                        if match is None:
                            return JSON_INVALID
                        token = (None, match) # HACK: shortcut to: (token[0], match, JSON_OBJECT if kind == u'{' else JSON_ARRAY)
                        just_closed = True

                token = InternalJsonGetToken(json, token[1]) # prepare next token
                kind = token[2]

            just_closed = False
            # Process coma if it can be present
            if not just_open:
                if kind == u',':
                    token = InternalJsonGetToken(json, token[1]) # load next token
                    if stk[-1] == u'[':
                        curkey += 1
                    continue

            if kind == u'}' and stk[-1] == u'{' or kind == u']' and stk[-1] == u'[':
                stk = stk[:-1]
                matchlvl -= 1
                if stk == []:
                    if InternalJsonGetToken(json, token[1])[2] is None:
                        break # Yay! end of job!
                    return JSON_INVALID # No yay - something at end of string
                just_closed = True
                token = InternalJsonGetToken(json, token[1])
                continue

            return JSON_INVALID


    except EInternalJsonInvalid:
        return JSON_INVALID

    if ret is None:
        return JSON_INVALID
    if ReturnsToken:
        return ret
    if ret[2] == JSON_STRING:
        return InternalJsonUnquote(json[ret[0]:ret[1]])
    if ret[2] in (JSON_NUMBER, JSON_OBJECT, JSON_ARRAY):
        return json[ret[0]:ret[1]]
    return ret[2] # JSON_TRUE, JSON_FALSE, JSON_NULL

def InternalJson2Elem(json):
    if json == u'': # checking this now lets us check for json[0] and json[-1] later
        return u''

    if json == u'null':
        return JSON_NULL

    if json == u'false':
        return JSON_FALSE

    if json == u'true':
        return JSON_TRUE

    match = (jsonnumbug_re if 6466 in Bugs else jsonnum_re).match(json)
    if match and match.end() == len(json):
        # HACK: Use our RE to know if the number is an integer
        if not match.group(1) and not match.group(2):
            # we have just digits with optional minus sign, i.e. an integer
            if len(json) > 11: # surely overflown
                if json[0] == u'-':
                    return -2147483648
                return 2147483647
            # a bit harder to test; we could check in ASCII to avoid conversion
            # to long in 32 bit systems, but it's probably not worth the effort
            elem = int(json)
            if elem > 2147483647:
                return 2147483647
            if elem < -2147483648:
                return -2147483648
            return elem
        return InternalTypecast(json, float, InList=False, f32=True)

    # Malformed strings are valid, e.g. "a\" (final \" is converted into a \)
    if json[0] == json[-1] == u'"' and json[1:2]: # the latter check ensures len(json) > 1
        return InternalJsonUnquote(json)

    return json

def llJson2List(json):
    json = fs(json)
    json = llStringTrim(json, 3) # STRING_TRIM

    if json == u'':
        return []

    if json[0] == u'[' and json[-1] == u']':
        # Array can of worms. Not all LSL quirks are implemented.
        ret = []
        token = InternalJsonGetTokenFull(json, 1)
        if token[2] == u']' and token[1] == len(json):
            return ret
        if token[2] == u':':
            return [JSON_INVALID]
        if token[2] == u',':
            ret.append(u'')
        else:
            ret.append(InternalJson2Elem(json[token[0]:token[1]]))
            token = InternalJsonGetTokenFull(json, token[1])
        while True:
            if token[2] == u']' and token[1] == len(json):
                break
            elif token[2] != u',':
                return [JSON_INVALID]
            token = InternalJsonGetTokenFull(json, token[1])
            if token[2] == u',' or token[2] == u']' and token[1] == len(json):
                ret.append(u'')
            else:
                if token[2] == u':':
                    return JSON_INVALID
                ret.append(InternalJson2Elem(json[token[0]:token[1]]))
                token = InternalJsonGetTokenFull(json, token[1])
        return ret

    if json[0] == u'{' and json[-1] == u'}':
        # Object can of worms. Worse than array. Not all LSL quirks are implemented.

        # Parse this grammar:
        # object: '{' complete_list incomplete_element '}' $
        # complete_list: <empty> | complete_list complete_element ','
        # complete_element: nonempty_string ':' value
        # incomplete_element: <empty> | value | string ':' value
        # string: '"' '"' | nonempty_string
        #
        # That allows:
        # {"a":1,"b":2,} # incomplete_element is empty
        # {"a":1,"b":2} # "b" is an incomplete_element
        # {2} # complete_list empty
        # {} # both empty
        # etc.

        ret = []
        token = InternalJsonGetTokenFull(json, 1)
        if token[2] == u'}' and token[1] == len(json):
            return ret
        if token[2] in (u':', u','):
            return [JSON_INVALID]

        while True:
            k = u''
            if token[2] == u'}' and token[1] == len(json):
                ret.append(k)
                ret.append(k)
                return ret
            if token[2] == JSON_STRING:
                colon = InternalJsonGetTokenFull(json, token[1])
                if colon[2] == u':':
                    k = InternalJsonUnquote(json[token[0]:token[1]])
                    token = InternalJsonGetTokenFull(json, colon[1])
            if token[2] in (u',', u':'):
                return [JSON_INVALID]
            ret.append(k)
            ret.append(InternalJson2Elem(json[token[0]:token[1]]))
            token = InternalJsonGetTokenFull(json, token[1])
            if token[2] == u'}' and token[1] == len(json):
                return ret
            if token[2] != u',' or k == u'':
                return [JSON_INVALID]
            token = InternalJsonGetTokenFull(json, token[1])

    return [InternalJson2Elem(json)]

def llJsonGetValue(json, lst):
    json = fs(json)
    lst = fl(lst)
    return InternalJsonFindValue(json, lst, ReturnsToken=False)

# llJsonSetValue was finally not implemented. This is a failed attempt
# at tackling it in the way that LSL does it.
'''def InternalJsonRecuriveSetValue(json, lst, val):
    # We give up and make it recursive

    if lst == []:
        if val == JSON_DELETE:
            return val
        return InternalElement2Json(val, ParseNumbers=True)

    ret = None
    lst0 = lst[0]
    tlst0 = type(lst0)
    if tlst0 == Key:
        tlst0 = unicode

    if val != JSON_DELETE:

        json = llStringTrim(json, 3) # STRING_TRIM
        if tlst0 == int and json[0:1] == u'[' and json[-1:] == u']':
            ret = []
            close = u']'
        if tlst0 == unicode and json[0:1] == u'{' and json[-1:] == u'}':
            ret = {}
            close = u'}'

    if ret is not None:
        if close: pass



def llJsonSetValue(json, lst, val):
    json = fs(json)
    lst = fl(lst)
    val = fs(val)
    if lst == []:
        # [] replaces the entire string no matter if it was invalid
        if val == JSON_DELETE:
            return val # this is a special case for SetValue with []
        return InternalElement2Json(val, ParseNumbers=True)
    # Needs to cope with JSON_APPEND, JSON_DELETE, lastindex+1.
    # Needs to do deep assignment.

    # Recursive works best here
    return InternalJsonRecursiveSetValue(json, lst, val)

    return u"----unimplemented----"
'''

def llJsonValueType(json, lst):
    json = fs(json)
    lst = fl(lst)
    ret = InternalJsonFindValue(json, lst, ReturnsToken=True)
    if ret == JSON_INVALID:
        return ret
    return ret[2]

def llList2Json(kind, lst):
    kind = fs(kind)
    lst = fl(lst)

    if kind == JSON_OBJECT:
        ret = u'{'
        if len(lst) & 1:
            return JSON_INVALID
        for i in xrange(0, len(lst), 2):
            if ret != u'{':
                ret += u','
            ret += InternalJsonQuote(lst[i]) + u':' + InternalElement2Json(lst[i+1], ParseNumbers=False)

        ret += u'}'

    elif kind == JSON_ARRAY:
        ret = u'['
        if lst:
            ret += InternalElement2Json(lst[0], ParseNumbers=False)
            for elem in lst[1:]:
                ret += u',' + InternalElement2Json(elem, ParseNumbers=False)
        ret += u']'

    else:
        ret = JSON_INVALID

    return ret



#!/usr/bin/env python
import subprocess

from setuptools import setup, find_packages
from setuptools.extension import Extension
from setuptools.command.build_ext import build_ext

try:
    from wheel.bdist_wheel import bdist_wheel
except ImportError:
    bdist_wheel = None


extra_cmds = {}


if bdist_wheel is not None:
    class BDistWheelABI3(bdist_wheel):
        def get_tag(self):
            python, abi, plat = super().get_tag()

            if python.startswith("cp"):
                # on CPython, our wheels are abi3 and compatible back to 3.8
                return "cp38", "abi3", plat

            return python, abi, plat

    extra_cmds["bdist_wheel"] = BDistWheelABI3


class AutobuildBuildExt(build_ext):
    """Build an extension that requires autobuild dependencies"""
    def build_extension(self, ext):
        autobuild_deps = getattr(ext, 'autobuild_deps', [])
        if autobuild_deps:
            subprocess.check_call(["autobuild", "install", "-A64", *autobuild_deps])
        build_ext.build_extension(self, ext)


class AutobuildExtension(Extension):
    def __init__(self, name, /, *, autobuild_deps=None, **kwargs):
        self.autobuild_deps = autobuild_deps or []
        super().__init__(name, **kwargs)


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='lummao',
    license='GPLv3',
    description='Toolkit for compiling and executing the Linden Scripting Language as Python',
    long_description=readme(),
    long_description_content_type="text/markdown",
    url='https://github.com/SaladDais/Lummao',
    author='Salad Dais',
    author_email='SaladDais@users.noreply.github.com',
    packages=find_packages(include=['lummao', 'lummao.*']),
    data_files=[],
    install_requires=['bytecode>=0.14,<1.0', 'aiohttp>=3.6,<4.0'],
    python_requires='>=3.8',
    zip_safe=False,
    use_scm_version=True,
    setup_requires=['setuptools_scm', 'autobuild<4'],
    tests_require=[
        "pytest",
        "pytest-cov",
        "wheel",
        "pytest-httpbin",
        "werkzeug<2.1",
    ],
    test_suite="tests",
    cmdclass={"build_ext": AutobuildBuildExt, **extra_cmds},
    ext_modules=[
        AutobuildExtension(
            "lummao._compiler",
            sources=["src/python_pass.cc", "src/json_ir_pass.cc", "src/compiler.cc"],
            define_macros=[("Py_LIMITED_API", "0x03080000")],
            libraries=["tailslide"],
            library_dirs=["build/packages/lib/release"],
            include_dirs=["build/packages/include"],
            extra_compile_args=["-std=c++17"],
            py_limited_api=True,
            autobuild_deps=["tailslide"],
        )
    ],
    entry_points={
        'console_scripts': {
            'lummao = lummao.cli:cli_main',
            'shellsl = lummao.shellsl:shellsl_main',
        }
    },
)



#!/usr/bin/env python
import asyncio
import sys
import uuid
from subprocess import Popen, PIPE

import lummao
from lummao import ScriptExtender, BaseLSLScript, TimerScriptExtender, DateTimeScriptExtender
from lummao.http import HTTPRequestScriptExtender

# All on one line so that we don't mess up line numbers for script errors
LSL_HEADER = rb"""
integer argc;
list argv;
integer cashquery; /* $? in bash */

system(string command) {}
string exec(list command_parts) {return "";}
""".replace(b"\n", b" ")


class ShelLSLExtender(ScriptExtender):
    script: lummao.BaseLSLScript

    def __init__(self, new_argv):
        super().__init__()
        self.argv = new_argv

    def extend_script(self, script: BaseLSLScript):
        super().extend_script(script)
        script.builtin_funcs["llOwnerSay"] = print
        script.builtin_funcs["llGetOwner"] = lambda: uuid.UUID(int=1)
        script.builtin_funcs["llLoadURL"] = lambda user, text, url: print("llLoadURL:", text, url)
        script.builtin_funcs["llResetScript"] = lambda: sys.exit(0)
        script.exec = self._exec_wrapper
        script.system = self._system_wrapper
        script.argc = len(self.argv)
        script.argv = self.argv

    async def _system_wrapper(self, command: str):
        p = Popen(command, shell=True)
        p.wait()
        self.script.cashquery = p.returncode

    async def _exec_wrapper(self, command: list):
        p = Popen(command, stdout=PIPE, stdin=PIPE)
        stdout, stderr = p.communicate()
        if stderr:
            print(stderr.decode("utf8"), file=sys.stderr)
        self.script.cashquery = p.returncode
        return stdout.decode("utf8").rstrip("\n")


def shellsl_main():
    with open(sys.argv[1], "rb") as f:
        lsl_bytes = f.read()
    # replace shebang line if present
    if lsl_bytes.startswith(b"#!"):
        lsl_bytes = b"\n" + b"\n".join(lsl_bytes.split(b"\n")[1:])
    script = lummao.compile_script(LSL_HEADER + lsl_bytes)

    # Enhance our script env with real HTTP, timer and shell support.
    shell_extender = ShelLSLExtender(sys.argv[1:])
    shell_extender.extend_script(script)
    http_extender = HTTPRequestScriptExtender()
    http_extender.extend_script(script)
    timer_extender = TimerScriptExtender()
    timer_extender.extend_script(script)
    datetime_extender = DateTimeScriptExtender()
    datetime_extender.extend_script(script)

    asyncio.run(script.execute_until_complete())


if __name__ == "__main__":
    shellsl_main()
